/*
 * This file is part of the YASMIN 8051 simulator project
 *
 * Copyright (c) 2000 by Gwenn Englebienne
 *
 * This software is free software; it is protected by the GPL. You can 
 * modify and/or redistribute it under the terms specified by the GNU
 * Public Licence version 2.
 *
 * Be advised that this software comes with no warranty at all, not
 * even the implied warranty of merchantability or fitness for a 
 * particular purpose. 
 *
 * For more information, you could look at the COPYING file, or have a 
 * look at the free software foundation's web site: http://www.fsf.org
 *
 */
 
/*
 * This software was written by Gwenn Englebienne on 28-29 may 1999.
 * All the source code you'll find here is protected by the GNU Public
 * License. In short, this means you have every right to use, hack, and
 * reuse all or parts of this code in other programs, as long as you make
 * the results public. For more information about the GPL, have a look at
 * http://www.fsf.org (free software foundation)
 *
 * To compile this code, you'll need the gtk libraries. I don't think you'll
 * need anything else (apart from the standard libraries, of course).
 *
 * Enjoy!
 */



void unknown(void)
{
     cout << "This function (" << instruction_name
	  << ") hasn't been implemented yet. Sorry for the inconvenience...\n";
     line_nr++;
}

void no_movc(void)
{
     cout << "\nBecause of the internal construction of this simulator,\n";
     cout << "the MOVC instruction isn't, and will probably never be "
	  << "implemented...\n\n";
     cout << "Any flames will be happilly sent to /dev/null. :->\n";
     line_nr++;
}

void jump (void)
{
     if (labels.contains(dest_name)) {
	  if (DEBUG_2)
	       cout << "Jumping to " << dest_name << ", line "
		    << labels[dest_name] << "\n";
	  line_nr = labels[dest_name];
     }
     else {
	  cout << "ERROR line " << line_nr << ": unknown label " << dest_name
	       <<endl;
	  exit(1);
     }
}

#define BIT 1
#define BYTE 0
#define WORD 456

void move (void)
{
     unsigned char dest_addr;
     char source_value;
     int direct = 1, source_type = 0, dest_type = 0;
     
     if (!strcmp(dest_name, "C")) {
	  dest_type = BIT;
	  dest_addr = sfr_info[dest_name].address;
     }
     else if (is_a_nr(dest_name)) {
	  dest_addr = asm_to_char (dest_name);
	  dest_type = BYTE;
     }
     else if (dest_name[0] == '@') {   // it's indirect
	  if (dest_name[1] == 'R' &&
	      (dest_name[2] == '0' || dest_name[2] == '1') &&
	      dest_name[3] == 0) {                      // Allow R0 and R1, nothing else
	       if (DEBUG_2)
		    cout << "Indirect dest addr: "
			 << (int) sfr_info[source_name].address
			 << ", addr2: "
			 << (int) da (sfr_info[source_name].address)
			 << ", value = "
			 << (int) ida (da (sfr_info[source_name].address))
			 << endl;
	       dest_addr = da(sfr_info[dest_name + 1].address);
	       dest_type = BYTE;
	       direct = 0;             // like I said, it's indirect
	  }
	  else
	       error("Indirect move only works with R0 and R1");
     } else if (sfr_info.contains(dest_name))
	  if (sfr_info[dest_name].type == byte) {
	       if (DEBUG)
		    cerr << "dest_addr = "
			 << (int)sfr_info[dest_name].address << endl;
	       dest_addr = sfr_info[dest_name].address;
	       dest_type = BYTE;
	  }
	  else {
	       dest_type = BIT;
	       dest_addr = sfr_info[dest_name].address;
	  }
     else if (!strcmp (dest_name, "DPTR"))  // It's a word, so let's avoid trouble: do it all here...
	  if (source_name[0] == '#') {
	       set_dptr (asm_to_int (source_name+1));
	       line_nr++;
	       return;                      // ... and bail out
	  }
	  else
	       error ("Invalid destination");
     
     if (!strcmp(source_name, "C")) {   // move from carry
	  source_type = BIT;
	  source_value = is_bit_set (sfr_info["C"].address);
     }
     else if (is_a_nr(source_name)) {    // move from byte address
	  source_value = da(asm_to_char (source_name));
	  source_type = BYTE;
     }
     else if (source_name[0] == '#') {       
	  source_value = asm_to_char(source_name + 1);
	  source_type = BYTE;
     }
     else if (source_name[0] == '@') {
	  if (source_name[1] == 'R' && (source_name[2] == '0' ||
					source_name[2] == '1') &&
	      source_name[3] == 0) {
	       if (DEBUG)
		    cout << "Indirect source addr: " << (int) sfr_info[source_name].address
			 << ", addr2: " << (int) da (sfr_info[source_name + 1].address)
			 << ", value = " << (int) ida (da (sfr_info[source_name+ 1].address)) << endl;
	       source_value = ida ( da (sfr_info[source_name +1 ].address));
	       source_type = BYTE;
	  }
	  else
	       error ("Indirect works only with R0 & R1");
     }
     else if (sfr_info.contains(source_name))
	  if (sfr_info[source_name].type == byte) {
	       source_value = da (sfr_info[source_name].address);
	       source_type = BYTE;
	  }
	  else {
	       source_type = BIT;
	       source_value = is_bit_set (sfr_info[source_name].address);
	  }
     else 
	  error ("Invalid source");
     
     if (source_type + dest_type == 0)     // byte --> byte
	  if (direct) {
	       da (dest_addr) = source_value;
	       if (DEBUG_2)
		    cout << "Direct move  ";
	  }
	  else 
	       ida (dest_addr) = source_value;
     else if (source_type + dest_type == 2) // bit --> bit
	  if (source_value)
	       set_bit(dest_addr);
	  else
	       clr_bit (dest_addr);
     else
	  error ("Wrong combination of operands for move!");
     
     zero = (source_value == 0);
     
     if (DEBUG_2) {
	  cout << "Moving " << source_name << " (" << int (source_value)
	       << ") to " << dest_name << " (" << setbase(16)
	       << (int)dest_addr << ")" << setbase(10) << "   => "
	       << (int)da(dest_addr) << endl;
	  cout << "dest_type = " << dest_type << ", source_type == "
	       << source_type << endl;
     }
     line_nr++;
}

void add (void)
{
     unsigned char dest_addr;
     char source_value;
     unsigned char negative = da (sfr_info["A"].address) < 0; // & 0x80;
     if (strcmp (dest_name, "A")) // => not equal to 'A'
	  error ("Addition needs accumulator as destination");
     else
	  dest_addr = sfr_info["A"].address;
     
     if (is_a_nr (source_name))
	  source_value = da (asm_to_char (source_name));
     else if (source_name[0] == 'R')
	  source_value = da (sfr_info[source_name].address);
     else if (source_name[0] == '#')
	  source_value = asm_to_char (source_name + 1);
     else if (source_name[0] == '@' && source_name[1] == 'R' &&
	      (source_name[2] == '0' || source_name[2] == '1') &&
	      source_name[3] == 0)
	  source_value = ida (da (sfr_info[source_name + 1].address));
     else
	  error ("invalid source");
     
     unsigned char psw_addr = sfr_info["PSW"].address;
     if (DEBUG) {
	  cerr << "Addition; source = " << int(source_value) << endl;
	  cerr << "          dest_addr = " << (int)dest_addr << " = "
	       << (int)da(dest_addr) << endl;
     }
     if (((da (dest_addr) & 0x0F) + (source_value & 0x0F)) & 0xF0)
	  da (psw_addr) |= AC;
     else
	  da (psw_addr) &= NOT_AC;
     
     da (dest_addr) += source_value;
     
     if (DEBUG)
	  cerr << "result : " << (int) da (dest_addr) << endl;
     if (negative && (da (dest_addr) >= 0)) 
	  da (psw_addr) |= CY;
     else
	  da (psw_addr) &= NOT_CY;
     
     if ((da (dest_addr) & 0x80) && (!negative))
	  da (psw_addr) |= OV;
     else
	  da (psw_addr) &= NOT_OV;

     zero = da (dest_addr) == 0;
     
     if (DEBUG)
	  cerr << "result : " << (int) da (dest_addr) << endl;
     if (DEBUG_2)
	  cout << "Adding " << source_name << " (" << (int)source_value
	       << ") to " << dest_name << endl;
     line_nr++;
}

void addc (void)
{
  unsigned char dest_addr;
  char source_value;
  int carry = (da (sfr_info["PSW"].address) & CY ) != 0;
  
  unsigned char negative = da (sfr_info["A"].address) < 0; // & 0x80;
  if (strcmp (dest_name, "A")) // => not equal to 'A'
    error ("Addition needs accumulator as destination");
  else
    dest_addr = sfr_info["A"].address;

  if (is_a_nr (source_name))
    source_value = da (atoi (source_name));
  else if (source_name[0] == 'R')
    source_value = da (sfr_info[source_name].address);
  else if (source_name[0] == '#')
    source_value = atoi (source_name + 1);
  else if (source_name[0] == '@' && source_name[1] == 'R' &&
	   (source_name[2] == '0' || source_name[2] == '1') &&
	   source_name[3] == 0)
    source_value = ida (da (sfr_info[source_name + 1].address));
  else
    error ("invalid source");

  unsigned char psw_addr = sfr_info["PSW"].address;
  if (((da (dest_addr) & 0x0F) + ((source_value + carry) & 0x0F)) & 0xF0)
    da (psw_addr) |= AC;
  else
    da (psw_addr) &= NOT_AC;

  da (dest_addr) += source_value + carry;

  if (negative && (da (dest_addr) >= 0)) 
    da (psw_addr) |= CY;
  else
    da (psw_addr) &= NOT_CY;
 
  if ((da (dest_addr) & 0x80) && (!negative))
    da (psw_addr) |= OV;
  else
    da (psw_addr) &= NOT_OV;

  zero = da (dest_addr) == 0;
  
  if (DEBUG_2)
    cout << "Adding " << source_name << " (" << (int)source_value << ") to "
	 << dest_name << endl;
  line_nr++;
}

void sub_borrow (void)
{

     int carry = (da (sfr_info["PSW"].address) & CY) != 0;
     if (DEBUG)
	  cout << "Carry = " << carry << endl;
     guint A = 1 << 8;
     
     unsigned char dest_addr;
     if (strcmp(dest_name, "A"))
	  error ("Substraction needs A as destination");
     dest_addr = sfr_info["A"].address;
     guchar source = source_val();

     A |= (unsigned char) da (dest_addr);

     if (DEBUG)
	  cout << "A = " << A << setbase(16) << " = " << A << setbase (10) << endl;

     A = A - source - carry;

     if (DEBUG)
	  cout << "A = " << A << setbase(16) << " = " << A << setbase(10) << endl;
     
     if (!(A & (1 << 8)))
	  da (sfr_info["PSW"].address) |= CY;
     else
	  da (sfr_info["PSW"].address) &= NOT_CY;
     if ((char) A > da (dest_addr))
	  da (sfr_info["PSW"].address) |= OV;
     else
	  da (sfr_info["PSW"].address) &= NOT_OV;
     
     da (dest_addr) = (char) A;
          
     zero = da (dest_addr) == 0;
     
     line_nr++; 
}

void push_on_stack (char c)
{
  da (sfr_info["SP"].address)++;
  ida ( da (sfr_info["SP"].address)) = c;
}

char& pop_from_stack (void)
{
  return ida ( da (sfr_info["SP"].address)-- );
}

void call (void)
{
  if (!labels.contains(dest_name))
    error ("Unknown call label");
  else {
    convert tmp;
    tmp.val = line_nr + 1;
    push_on_stack (tmp.split.low);
    push_on_stack (tmp.split.high);
    line_nr = labels[dest_name];
  }
}

void ret (void)
{
  convert tmp;
  tmp.split.high = pop_from_stack();
  tmp.split.low = pop_from_stack();
  line_nr = tmp.val;
}   

void reti (void)
{
  convert tmp;
  da (sfr_info["PSW"].address) = pop_from_stack();
  tmp.split.high = pop_from_stack();
  tmp.split.low = pop_from_stack();
  switch (interrupt_stack.pop()) {
  case timer_0:
       clr_bit (sfr_info["TF0"].address);
       break;
  case timer_1:
       clr_bit (sfr_info["TF1"].address);
       break;
  }
  line_nr = tmp.val;
}

void jump_on_zero (void)
{
  if (labels.contains (dest_name))
    if (zero)
      line_nr = labels[dest_name];
    else
      line_nr++;
  else
    error ("Label unknown");
}

void jump_not_zero (void)
{
  if (labels.contains (dest_name))
    if (!zero)
      line_nr = labels[dest_name];
    else
      line_nr++;
  else
    error ("Label unknown");
}

void djnz (void)
{
  unsigned char dest_addr;
  if (is_a_nr (dest_name))
    dest_addr = asm_to_char (dest_name);
  else if (sfr_info.contains(dest_name))
    dest_addr = sfr_info[dest_name].address;

  if (labels.contains(source_name) || !strcmp (source_name, "$")) {
    da (dest_addr)--;
    zero = da (dest_addr) == 0;
    if (!zero)
      if (strcmp (source_name, "$"))
	line_nr = labels[source_name];
      else;
    else
      line_nr++;
  } else
    error ("Unknown label");
}

void inc (void)
{
  unsigned char dest_addr;
  int direct = 1;
  
  if (is_a_nr(dest_name))
    dest_addr = asm_to_char (dest_name);
  else if (dest_name[0] == '@' &&
	   dest_name[1] == 'R' &&
	   (dest_name[2] == '0' || dest_name[2] == '1') &&
	   dest_name[3] == 0)
    dest_addr = da (sfr_info[dest_name + 1].address), direct = 0;
  else if (sfr_info.contains(dest_name) && sfr_info[dest_name].type == byte)
    dest_addr = sfr_info[dest_name].address;
  else if (!strcmp (dest_name, "DPTR")) {
    set_dptr (get_dptr () + 1);
    line_nr++;
    return;
  }
  else
    error ("Unkown operand.");

  if (direct)
    da (dest_addr)++;
  else
    ida (dest_addr)++;
  line_nr++;
}   

void dec (void)
{
  unsigned char dest_addr;
  int direct = 1;
  
  if (is_a_nr(dest_name))
    dest_addr = asm_to_char (dest_name);
  else if (dest_name[0] == '@' &&
	   dest_name[1] == 'R' &&
	   (dest_name[2] == '0' || dest_name[2] == '1'))
    dest_addr = da (sfr_info[dest_name + 1].address), direct = 0;
  else if (s)h
